type image = {
	img : (int*int*int) vect vect;
	energ : float vect vect;
	lum : float vect vect;
	mutable x : int;
	mutable y : int	
};;

(* 1 *)
let calcul_lum img =
  let x = img.x in
  let y = img.y in
    for i = 0 to x-1 do
      for j = 0 to y-1 do  
	let (r, v, b) = img.img.(i).(j) in
	  img.lum.(i).(j) <- 0.3*.(float_of_int r) +. 0.6*.(float_of_int v) +. 0.1*.(float_of_int b)
      done;
    done;;
	    
(* 2 *)
let calcul_energie img fonc =
  let x = img.x in
  let y = img.y in
    for i=0 to x-1 do
      for j=0 to y-1 do
	img.energ.(i).(j) <- fonc img i j
      done;
    done;;

  
type chemin == (int*int) list;;

(* 3 *)

(* On retire juste un chemin de l'image, et l'on en profite 
pour faire la mme chose sur la luminosit au passage.
En revanche, on ne peut procder ainsi avec l'nergie, puisque
celle-ci dpend du voisinage et qu'il est modifi le long du chemin *)

let reducev img (path:chemin) =
  let y = img.y in
  let rec rempli = function
    |[] -> ()
    |(i, j)::l ->
       for k = j to y-2 do
  	 img.img.(i).(k) <- img.img.(i).(k+1);
	 (* On rajuste la luminosit *)
  	 img.lum.(i).(k) <- img.lum.(i).(k+1);
       done;
  	rempli l
  in
    img.y <- img.y-1;
    rempli path;;
      

(* 4 *)

(* Le cout du chemin est celui du min des c(n,j), 0<=j<p *)

(* 5 *)

(* c(i,j) = e(i,j) + min c(i-1,j-1),c(i-1,j),c(i-1,j+1) *)
(* c(0,j) = e(0,j) *)
(* c(i,0) = e(i,0) + min c(i-1,0),c(i-1,1) *)
(* c(i,n-1) = e(i,n-1) + min c(i-1,n-2),c(i-1,n-1) *)


(* 6 *)

(* Une premire version, o l'on stocke brutalement le chemin 
tout du long de la prog dyn *)


(* De petites fonctions auxiliaires bien utiles pour allger le code *)
let min2 c1 c2 =
  let a,_=c1 and  b,_ =c2 in
  if a<b
  then
    c1
  else c2;;

let min3 c1 c2 c3 = 
  let a,_ = c1 and b,_ = c2 and c,_ = c3 in
  if a<b
  then
    if a<c then c1
    else c3
  else 
    if b<c then c2
  else c3;;

let minimum t =
  let l = vect_length t in
  let im = ref 0 in
    for i = 0 to l-1 do
      if (fst t.(i)) < (fst t.(!im)) then im := i;
    done;
    t.(!im);;

(* La fonction en elle-mme *)

let cheminv img = 
  let y = img.y in
  let x = img.x in
  let t = make_matrix x y (0., []) in 
  let m = img.energ in
    (* Premiere ligne *)
    for j = 0 to y-1 do
      t.(0).(j) <- m.(0).(j), [0, j];
    done;
    
    (* Les autres de proches en proches *)
    for i = 1 to x-1 do
      (* Premiere colonne *)
      let s,d = min2 (t.(i-1).(0)) t.(i-1).(1) in
	t.(i).(0) <- (s+.m.(i).(0), (i,0)::d);
	
	(* Derniere colonne *)
	let s,d = min2 t.(i-1).(y-2) t.(i-1).(y-1) in
	  t.(i).(y-1) <- (s+.m.(i).(0), (i,y-1)::d);
	  
	  (* Les autres *)
	  for j=1 to y-2 do
	    let s,d=min3  t.(i-1).(j-1)  t.(i-1).(j)  t.(i-1).(j+1) in 
	      t.(i).(j) <- (s+.m.(i).(j), (i,j)::d)
	  done;
	done;
    let (s, path) = minimum t.(x-1) in (path:chemin);;


(* Une seconde version, un peu plus fine, o l'on ne garde que l'antcdent, et o l'on remonte *)

let read_path t start= 
  let rec aux l = function
    |0,k-> ((0,k)::l:chemin)
    |i,j -> let _,(_,f) = t.(i).(j) in aux ((i,j)::l) (i-1,f)
  in aux [] start;;

let min2 a b =
  if a<b
  then
    a,0
  else b,1;;

let min3 a b c =
  if a<b
  then
    if a<c then a,-1
    else c,1
  else 
    if b<c then b,0
  else c,1;;


let cheminv img = 
  let y = img.y in
  let x = img.x in
  let t = make_matrix x y (0., (0,0)) in 
  let m = img.energ in
    (* Premiere ligne *)
    for j = 0 to y-1 do
      t.(0).(j) <- m.(0).(j), (0, -1);
    done;
    
    (* Les autres de proches en proches *)
    for i = 1 to x-1 do
      (* Premiere colonne *)
      let (s1, l1) = t.(i-1).(0) and (s2, l2) = t.(i-1).(1) in
      let s,d = min2 s1 s2 in
	t.(i).(0) <- (s+.m.(i).(0), (i-1,d));
	(* Derniere colonne *)
	let (s1, l1) = t.(i-1).(y-2) and (s2, l2) = t.(i-1).(y-1) in
	let s,d = min2 s1 s2 in
	  t.(i).(y-1) <- (s+.m.(i).(0), (i-1,y-2+d));
	  (* Les autres *)
	  for j=1 to y-2 do
	    let (s1, l1) = t.(i-1).(j-1) and (s2, l2) = t.(i-1).(j) and (s3, l3) = t.(i-1).(j+1) in
	    let s,d=min3 s1 s2 s3 in 
	      t.(i).(j) <- (s+.m.(i).(j), (i-1,j+d))
	  done;
	done;
	      let (s, start) = minimum t.(x-1) in read_path t start;;
  
      



(* 7 *)
type fe ==(image -> int -> int -> float);;

let seamcarving img nombre (fonc:fe) =
  for i=0 to nombre do
    (* bien penser  recalculer la matrice d'nergie *)
    calcul_energie img fonc;
    let chemin = cheminv img in 
      reducev img chemin
  done;
  img;;
  

(* Une fonction classique de calcul d'nergie, base sur un gradient *)
let carre x = x*.x;;

let energie_discret image i j =
  let aux i j =
    let xm = image.x-1 in
    let ym = image.y-1 in 
    let m = image.lum in
      match (i,j) with
	|0,0-> 2.*.(m.(0).(1)-.m.(0).(0)),2.*.(m.(1).(0)-.m.(0).(0))
	|0,b when b=ym -> 2.*.(m.(0).(ym)-.m.(0).(ym-1)),2.*.(m.(1).(ym)-.m.(0).(ym))
	|0,b -> (m.(0).(b+1)-.m.(0).(b-1)),2.*.(m.(1).(b)-.m.(0).(b))	  
	|a,0 when a=xm-> 2.*.(m.(xm).(1)-.m.(xm).(0)),2.*.(m.(xm-1).(0)-.m.(xm).(0))
	|a,0 -> (m.(a).(1)-.m.(xm).(0)),(m.(a-1).(0)-.m.(a+1).(0))	    
	|a,b when (a,b)=(xm,ym)-> 2.*.(m.(xm).(ym)-.m.(xm).(ym-1)),2.*.(m.(xm).(ym)-.m.(xm-1).(ym))
	|a,b when a=xm-> m.(xm).(b+1)-.m.(xm).(b-1),2.*.(m.(xm).(b)-.m.(xm-1).(b))
	|a,b when b=ym-> 2.*.(m.(a).(ym)-.m.(a).(ym-1)),(m.(a+1).(ym)-.m.(a-1).(ym))
	|u,v-> m.(i+1).(j)-.m.(i-1).(j), m.(i).(j+1)-.m.(i).(j-1) 
  in
  let dh,dv=aux i j 
  in   
    sqrt((carre dh)+.(carre dv));;


(* Outils de lecture/criture pour les images en PPM *)
 
exception Wrong_type;;
  
let input_ws in_c = 
  let s = ref "" in
  let c = ref (input_char in_c) in
    try
      while (neq_string (string_of_char !c) "\n" ) && (neq_string (string_of_char !c) "\t") && (neq_string (string_of_char !c) " ") do
	s := (!s)^(make_string 1 (!c));
	c := input_char in_c; 
      done;
      !s;
    with
      |End_of_file -> !s;;

(* Lecture header *)
let read_hd f =
  let format = input_ws f in
  let largeur = int_of_string (input_ws f) in
  let hauteur = int_of_string (input_ws f) in
  let niveau_gris = int_of_string (input_ws f) in
    format, largeur, hauteur, niveau_gris;;

(* Cration de la matrice *)
let matrix_gris largeur hauteur input =
  let m = make_matrix hauteur largeur (0, 0, 0) in
    for i=0 to hauteur-1 do
      for j=0 to largeur-1 do
	let g = input_byte input in
	  m.(i).(j) <- (g, g, g)
      done;
      done;
    m;;

let matrix_color largeur hauteur input =
  let m = make_matrix hauteur largeur (0, 0, 0) in
    for i=0 to hauteur-1 do
      for j=0 to largeur-1 do
	let (b,v,r) = (input_byte input, input_byte input, input_byte input) in
	  m.(i).(j) <- (r, v, b)
      done;
      done;
	  m;;
	    
let generate_matrix fichier =		  
  let input = open_in_bin fichier in
  let (version, largeur, hauteur, n_g) = read_hd input in
  let m =
    match version with
      |"P5" -> matrix_gris largeur hauteur input
      |"P6" -> matrix_color largeur hauteur input
      |_ -> raise Wrong_type    
  in
  let r = {img = m; energ = make_matrix hauteur largeur 0.; lum = make_matrix hauteur largeur 0.; x = hauteur; y = largeur} in
    calcul_lum r;
    print_string "coucou";
    calcul_energie r energie_discret; 
    r;;
      

(* Export en PPM *)
let save_matrix img w =
  let x = img.x in
  let y = img.y in
  let f = open_out_bin w in
    output_string f ("P6\n"^(string_of_int y)^" "^(string_of_int x)^"\n255\n");
    
    for i = 0 to x-1 do
      for j = 0 to y-1 do
	let (r, v, b) = img.img.(i).(j) in
	  output_byte f r;
	  output_byte f v;
	  output_byte f b;
      done;
      done;
      close_out f;;


(* On teste ! *)
let file="images/arbre.ppm";;
let image=generate_matrix file;; 
 
(* Pour colorier un chemin en rouge *)
let rec color_red img = function
  |[] -> ()
  |(i,j)::l -> img.img.(i).(j)<- (255,0,0);color_red img l;;

let path=cheminv image;; 
color_red image path;; 
save_matrix image "images/arbre_red.ppm";; 
let cut = seamcarving image 150 energie_discret;;
save_matrix cut "images/arbre_cut.ppm";;



(* 8 *)
(* L'ide est ici de ne recalculer l'nergie que le long du chemin supprim *)
let calcul_next_energie_v img path fonc =
  let y = img.y+1 in
  let rec rempli = function
    |[] -> ()
    |(i, j)::l ->
       for k = (max 0 (j-1)) to y-2 do
	 if k=j-1 then img.energ.(i).(k) <- fonc img i k
	 else if k=j then img.energ.(i).(k) <- fonc img i k
	 else img.energ.(i).(k) <- img.energ.(i).(k+1);
       done;
	rempli l
  in
    rempli path;;
      

						
 						;;
let reducev img (path:chemin) (fonc:fe) =
  let y = img.y in
  let rec rempli = function
    |[] -> ()
    |(i, j)::l ->
       for k = j to y-2 do
  	 img.img.(i).(k) <- img.img.(i).(k+1);
  	 img.lum.(i).(k) <- img.lum.(i).(k+1);
       done;
  	rempli l
  in
    img.y <- img.y-1;
    calcul_next_energie_v img path fonc;
    rempli path;;

let seamcarving img nombre (fonc:fe) =
  for i=0 to nombre do
    reducev img (cheminv img) fonc
  done;
  img;;


(* 9 *)

(* Cette question est assez ouverte. Si l'on duplique  chaque fois le chemin de plus faible
nergie, on va dupliquer  chaque fois le mme, et c'est assez vite moche. 
L'ide va tre au contraire de slectionner un certain nombre de chemins d'un coup, 
et des les dupliquer. 
Il faut pour cela se fixer des quotas  ne pas dpasser 
(pas plus de chemins d'un coup qu'une certaine proportion de la largeur de l'image), 
et ventuellement regarder le problme des chemins non spars 
(si l'on prend deux chemins qui se croisent, un pixels sera doubler, et ainsi de suite). 
Une bonne mthode consiste de fait  simuler la suppression de n chemins, 
puis de copier ces n chemins.
Pour plus de dtails, je vous conseille de regarder l'article originel sur le SeamCarving *)
